/* ReSCAL - Simulation of the very specific processes of the models
 *
 * Copyright (C) 2021
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * Code based on dissol program,
 * by Eduardo Sepulveda <edo@espci.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <memory.h>
#include <math.h>
#include <time.h>

#include "defs.h"
#include "macros.h"
#include "param.h"
#include "space.h"
#include "surface.h"
#include "simul.h"
#include "simul_models.h"
#include "schedul.h"
#include "trace.h"
#include "lgca.h"
#include "callbacks.h"
#include "interface.h"
#include "view.h"

extern unsigned char opt_nv;
extern int H, L, D, HL, HLD;       // les dimensions de la terre
extern int Ncel[];                       // nombre de cellules par type
extern int flag_csp_altered; /// cellular space has been altered
extern char *rot_map;        // periodic mapping of the rotating space
extern int use_lgca;
#ifdef LGCA
extern int lgca_speedup;
extern int lgca_change_visc;
extern int lgca_iter;    //nombre de cycles de collisions
extern char *nom_fic_mvt;
extern VelSpace Vsv;
#ifdef CGV
extern float grdvc_max; //seuil max pour le controle de l'erosion en fonction du gradient de vitesse
extern float grdvc_min; //seuil min pour le controle de l'erosion en fonction du gradient de vitesse
#ifdef CELL_COLOR
extern float grdvc_max_col; //seuil max pour le controle de l'erosion des grains colores en fonction du gradient de vitesse
extern float grdvc_min_col; //seuil min pour le controle de l'erosion des grains colores en fonction du gradient de vitesse
#endif
#endif
#endif
extern float lambda_A;
extern float lambda_A_unstable;
#ifdef CELL_COLOR
//extern float color_coef;         // coefficient associe a la callback check_color
#endif

extern int vdir_mode; //display mode of the current orientation
extern unsigned char reorient_flag;

extern WindRose wind_rose;

extern double csp_time;                  // time variable for the cellular space
extern double stop_time;                 // stopping time
extern int flag_save_auto; /// save data before each rotation

extern int boundary;  //boundary conditions
extern int pbc_mode; //periodic boundary conditions

#ifdef TIME_SCALE
double real_time=0.0;       // real time with time scale
#endif

int ava_mode = AVA_SYNC;  //mode d'avalanches
char *ava_mode_str = NULL;
int ava_trans = 0;        //transitions d'avalanches
int ava_h_lim = 1; //10;	// hauteur limite avant avalanche
int ava_nb_cel_max = 1;  // nb de cellules qui tombent simultanement
float ava_delay = 0.0; //delay between avalanches
float ava_duration = 1.0; //default value for DUN model
//float ava_delay_long = 0.0; //8000.0;
float ava_angle = 0.0; //angle of avalanches (degrees)
float ava_angle_col = 0.0; //angle of avalanches (degrees) for colored grains
float ava_angle_stable = 0.0;
float ava_angle_unstable = 0.0;
float ava_angle_stable_col = 0.0;
float ava_angle_unstable_col = 0.0;
int ava_angle_alti_min = 0;
float lambda_A_stable = 0.0;  //float time_0; //time scale

int ava_upwind = 1;
int ava_accu = AVA_AA;  //accuracy of avalanches
char *ava_accu_str = NULL;
int mode_pat;

#ifdef SURRECTIONS
float sur_delay = 0.0;  //10000.0;              // delai entre deux surrections successives (0 = pas de surrection)
int sur_mode = 0;                // type de surrection (uniforme ou locale)
char *sur_mode_str=NULL;
#endif

int rot_mode = ROT_MODE_OVERLAP;  //default rotation mode (assuming no rotation)
#ifdef ROTATIONS
float rot_angle = 0.0;
float rot_delay = 0.0;
char *rot_mode_str=NULL;
#endif
#ifdef CENTERING_AUTO
float centering_delay = 0.0;
#endif
int flag_restore_template = 0;
char *csp_template_str = NULL; //value of CSP_template parameter
int csp_template_type = CSP_CUSTOM; //CSP template

#ifdef LGCA
int init_mvt_nb_cycles = -1;  //nombre de cycles de stabilisation du flux au depart
int mvt_nb_cycles = -1;  //nombre de cycles de stabilisation (lgca)
double lgca_delay = 0.0; //delai entre deux cycles de gaz sur reseau
int lgca_ready = 0;
int flag_lgca_reset = 0;
#endif

char *wind_data_filename = NULL;
int wind_angle_is_set = 0;
#ifdef CGV
char *qsat_data_filename = NULL; //"real_data/PDF.data";
#endif
#ifdef TIME_SCALE
char *phys_prop_filename = NULL; //"real_data/desert_earth.prop";
float time_scale = 1.0; //t_0
#endif


#ifdef LGCA
void flow_stabilization(int );
#endif
void simul_wind_rose(int, float*, float* );
double simul_wind_data();
double get_next_time_threshold();

void params_simul_models()
{

#ifdef AVALANCHES
  param_family("AVALANCHES", "Avalanches parameters");
  //parameter("Ava_mode", "mode of avalanches (NONE|SYNC|TRANS|PROPAG) - SYNC by default", &ava_mode_str, PARAM_STRING, "AVALANCHES");
  parameter("Ava_mode", "mode of avalanches (NONE|SYNC|TRANS) - SYNC by default", &ava_mode_str, PARAM_STRING, "AVALANCHES");
  parameter("Lambda_A", "rate of avalanches in TRANS mode", &lambda_A, PARAM_FLOAT, "AVALANCHES");
  parameter("Lambda_A_unstable", "unstable avalanches rate in TRANS mode", &lambda_A_unstable, PARAM_FLOAT, "AVALANCHES");
  parameter("Ava_h_lim", "height limit for avalanches - 1 by default", &ava_h_lim, PARAM_INT, "AVALANCHES");
  parameter("Ava_nb_cel", "number of cells falling together in SYNC mode - 1 by default", &ava_nb_cel_max, PARAM_INT, "AVALANCHES");
  parameter("Ava_delay", "delay between avalanches (t0) - optional", &ava_delay, PARAM_FLOAT, "AVALANCHES");
  parameter("Ava_angle", "angle of repose (degrees) - optional", &ava_angle, PARAM_FLOAT, "AVALANCHES");
  parameter("Ava_angle_alti_min", "minimal altitude for estimating the slope in avalanches - 0 by default", &ava_angle_alti_min, PARAM_INT, "AVALANCHES");
  parameter("Ava_duration", "duration of avalanches (t0) - optional", &ava_duration, PARAM_FLOAT, "AVALANCHES");
  parameter("Ava_angle_unstable", "unstable angle of repose (degrees) - optional", &ava_angle_unstable, PARAM_FLOAT, "AVALANCHES");
  parameter("Ava_upwind", "enable upwind avalanches in TRANS mode (YES|NO) - YES by default", &ava_upwind, PARAM_BOOLEAN, "AVALANCHES");
  //parameter_hidden("Ava_anisotropy", "Ava_anisotropy parameter is obsolete - Please use Ava_upwind parameter instead !", &ava_upwind, PARAM_INT, "AVALANCHES");
  parameter("Ava_accuracy", "level of accuracy in avalanches (AVERAGE|HIGH) - AVERAGE by default", &ava_accu_str, PARAM_STRING, "AVALANCHES");
#ifdef CELL_COLOR
  parameter("Ava_angle_col", "angle of repose for white grains (degrees) - optional", &ava_angle_col, PARAM_FLOAT, "COLOR");
  parameter("Ava_angle_unstable_col", "unstable angle of repose for white grains (degrees) - optional", &ava_angle_unstable_col, PARAM_FLOAT, "COLOR");
#endif
#endif //AVALANCHES

/*
#ifdef CELL_COLOR
  parameter("Color_coef", "control coefficient of colored grains - optional", &color_coef, PARAM_FLOAT, "COLOR");
#endif
*/

#ifdef SURRECTIONS
  param_family("SURRECTION", "Surrections parameters");
  parameter("Sur_mode", "surrection mode (UNIFORM|CONE|TECTO)", &sur_mode_str, PARAM_STRING, "SURRECTION");
  parameter("Sur_delay", "surrection delay (t0)", &sur_delay, PARAM_FLOAT, "SURRECTION");
#endif

#ifdef ROTATIONS
  param_family("ROTATION", "Rotations parameters");
  parameter("Rot_mode", "rotation mode (FULL|DISK|OVERLAP) - optional", &rot_mode_str, PARAM_STRING, "ROTATION");
  parameter("Rot_angle", "rotation angle (degrees) - optional", &rot_angle, PARAM_FLOAT, "ROTATION");
  parameter("Rot_delay", "rotation delay (t0) - optional", &rot_delay, PARAM_FLOAT, "ROTATION");
#ifdef LGCA
  parameter("Lgca_nb_cycles", "number of cycles for flow stabilization - optional", &mvt_nb_cycles, PARAM_INT, "LGCA");
  //parameter("Rot_ncycl", "number of flow cycles without transitions after rotation - optional", &rot_mvt_nb_cycles, PARAM_INT, "LGCA");
#endif
  parameter("Restore_template", "restore circular source after rotation (YES|NO) - NO by default", &flag_restore_template, PARAM_BOOLEAN, "ROTATION");
  parameter("Csp_template", "CSP template for the generation of the initial cellular space", &csp_template_str, PARAM_STRING,"GENESIS");
#endif

#ifdef CENTERING_AUTO
  parameter("Centering_delay", "automatic centering delay (t0) - optional", &centering_delay, PARAM_FLOAT, "");
#endif

#ifdef LGCA
  parameter_hidden("Init_ncycl", "initial number of flow cycles without transitions - optional", &mvt_nb_cycles, PARAM_INT, "LGCA"); /// obsolete
  parameter("Lgca_delay", "time delay between flow cycles (t0) - optional", &lgca_delay, PARAM_DOUBLE, "LGCA");
#endif

  param_family("REAL", "Real-space parameters");
#ifdef CGV
//  parameter("Grdvc", "critical shear stress for relationship between erosion rate and shear stress - optional", &grdvc, PARAM_FLOAT);
  parameter("Tau_min", "min shear stress value in the linear relationship between erosion rate and shear stress - optional", &grdvc_min, PARAM_FLOAT, "MODEL");
  parameter("Tau_max", "max shear stress value in the linear relationship between erosion rate and shear stress - optional", &grdvc_max, PARAM_FLOAT, "MODEL");
  parameter_hidden("Grdvc_min", "Grdvc_min parameter is obsolete - Please use Tau_min parameter instead !", &grdvc_min, PARAM_FLOAT, "MODEL");
  parameter_hidden("Grdvc_max", "Grdvc_max parameter is obsolete - Please use Tau_max parameter instead !", &grdvc_max, PARAM_FLOAT, "MODEL");
#ifdef CELL_COLOR
  parameter("Tau_min_col", "min shear stress value for white grains in the linear relationship between erosion rate and shear stress - optional", &grdvc_min_col, PARAM_FLOAT, "COLOR");
  parameter("Tau_max_col", "max shear stress value for white grains in the linear relationship between erosion rate and shear stress - optional", &grdvc_max_col, PARAM_FLOAT, "COLOR");
  parameter_hidden("Grdvc_min_col", "Grdvc_min_col parameter is obsolete - Please use Tau_min_col parameter instead !", &grdvc_min_col, PARAM_FLOAT, "COLOR");
  parameter_hidden("Grdvc_max_col", "Grdvc_max_col parameter is obsolete - Please use Tau_max_col parameter instead !", &grdvc_max_col, PARAM_FLOAT, "COLOR");
#endif
  parameter("Qsat_file", "real PDF data between Qsat and tau1 - optional", &qsat_data_filename, PARAM_STRING, "REAL");
#endif //CGV
#ifdef WIND_DATA
  parameter("Wind_file", "filename for wind data - optional", &wind_data_filename, PARAM_STRING, "REAL");
#endif
#ifdef TIME_SCALE
  parameter("Phys_prop_file", "filename for physical parameters - optional", &phys_prop_filename, PARAM_STRING, "REAL");
#endif
}

void parse_simul_models()
{
  if (ava_mode_str){
    if (!strcmp(ava_mode_str, "NONE")){
      ava_mode = AVA_NONE;
    }
    else if (!strcmp(ava_mode_str, "SYNC")){
      ava_mode = AVA_SYNC;
    }
    else if (!strcmp(ava_mode_str, "TRANS")){
      ava_mode = AVA_TRANS;
    }
    else if (!strcmp(ava_mode_str, "PROPAG")){
      ava_mode = AVA_PROPAG;
      WarnPrintf("WARNING: obsolete mode of avalanches (%s)\n", ava_mode_str);
    }
    else{
      ErrPrintf("Incorrect value for Ava_mode: %s\n", ava_mode_str);
      exit(-2);
    }
  }
  if (ava_accu_str){
    if (!strcmp(ava_accu_str, "AVERAGE")){
      ava_accu = AVA_AA;
    }
    else if (!strcmp(ava_accu_str, "HIGH")){
      ava_accu = AVA_HA;
    }
    else{
      ErrPrintf("Incorrect value for Ava_accuracy: %s\n", ava_accu_str);
      exit(-2);
    }
  }
#ifdef SURRECTIONS
  if (sur_mode_str){
    if (!strcmp(sur_mode_str, "UNIFORM")){
      sur_mode = SUR_MODE_UNIFORM;
    }
    else if (!strcmp(sur_mode_str, "CONE")){
      sur_mode = SUR_MODE_CONE;
    }
    else if (!strcmp(sur_mode_str, "TECTO")){
      sur_mode = SUR_MODE_TECTO;
    }
    else{
      ErrPrintf("Incorrect value for Sur_mode: %s\n", sur_mode_str);
      exit(-2);
    }
  }
#endif
#ifdef ROTATIONS
  if (rot_mode_str){
    if (!strcmp(rot_mode_str, "DISK") || !strcmp(rot_mode_str, "TABLE")){
      rot_mode = ROT_MODE_DISK;
    }
    else if (!strcmp(rot_mode_str, "OVERLAP") || !strcmp(rot_mode_str, "CYCLE")){
      rot_mode = ROT_MODE_OVERLAP;
    }
    else if (!strcmp(rot_mode_str, "FULL") || !strcmp(rot_mode_str, "SQUARE")){
      rot_mode = ROT_MODE_FULL;
    }
    else{
      ErrPrintf("Incorrect value for Rot_mode: %s\n", rot_mode_str);
      exit(-2);
    }
  }
  if ((rot_mode == ROT_MODE_FULL) && (boundary == BC_REINJECTION)){
    ErrPrintf("Incorrect value for Rot_mode: FULL, not compatible with REINJECTION boundary conditions\n");
    exit(-2);
  }
#ifdef MODEL_DUN
  if (csp_template_str){
    if (!strncmp(csp_template_str, "SRC_DISK_CEIL",13)){
      csp_template_type = CSP_SRC_DISK_CEIL;
    }
    if (!strncmp(csp_template_str, "SRC_RECT",8)){
      csp_template_type = CSP_SRC_RECT;
    }
    /// other CSP types not supported yet
    LogPrintf("csp_template_type = %d\n", csp_template_type);
  }
#endif
#endif

  if (param_is_set("Init_ncycl")) WarnPrintf("WARNING: Init_ncycl is obsolete and has been replaced by Lgca_nb_cycles\n");

  ava_trans = (ava_mode == AVA_TRANS) || (ava_mode == AVA_PROPAG);
}


#ifdef TIME_SCALE
extern float lambda_C;
extern float lambda_T;

//#ifdef CGV
void simul_time_scale(float wind_value)
{
  FILE *fp;
  static int nb_rows;
  static float *tau1_values; //=grdvc_min
  static float *erosion_rates;
  static float *qsat_ratios;
  static float kappa; //Karman constant
  static float z_0; //roughness length
  static float z; //height of observation (velovity gradient)
  static float rho_air; //density of air
  static float rho_sand; //density of sand
  static float g; //gravity constant
  static float ut; //threshold velocity (erosion)
  static float d; //diameter of grain
  static float l_0; //length scale
  static char start=1;
  int i;

  if (!phys_prop_filename) return;

  if (start){
    //read physical parameters file
    LogPrintf("simul_time_scale: read physical parameters file %s\n", phys_prop_filename);

    if( (fp = fopen(phys_prop_filename,"r")) == NULL ) {
      ErrPrintf("ERROR: cannot open file %s\n",phys_prop_filename);
      exit(-4);
    }

    if (fscanf(fp,"kappa=%f\n",&kappa) != 1){
      ErrPrintf("ERROR: cannot read parameter kappa in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("kappa=%f\n", kappa);

    if (fscanf(fp,"z_0=%f\n",&z_0) != 1){
      ErrPrintf("ERROR: cannot read parameter z_0 in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("z_0=%f\n", z_0);

    if (fscanf(fp,"z=%f\n",&z) != 1){
      ErrPrintf("ERROR: cannot read parameter z in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("z=%f\n", z);

    if (fscanf(fp,"rhoair=%f\n",&rho_air) != 1){
      ErrPrintf("ERROR: cannot read parameter rho_air in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("rhoair=%f\n", rho_air);

    if (fscanf(fp,"rhosand=%f\n",&rho_sand) != 1){
      ErrPrintf("ERROR: cannot read parameter rho_sand in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("rhosand=%f\n", rho_sand);

    if (fscanf(fp,"g=%f\n",&g) != 1){
      ErrPrintf("ERROR: cannot read parameter g in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("g=%f\n", g);


    if (fscanf(fp,"d=%f\n",&d) != 1){
      ErrPrintf("ERROR: cannot read parameter d in %s\n",phys_prop_filename);
      exit(-4);
    }
    LogPrintf("d=%f\n", d);


    fclose(fp);

    ut = 0.1*sqrt((rho_sand/rho_air)*g*d);
    LogPrintf("ut=%f\n", ut);

    l_0=5*(rho_sand/rho_air)*d/4.;
    LogPrintf("l_0=%f\n", l_0);

    //read qsat data file
    LogPrintf("simul_time_scale: read qsat data file %s\n", qsat_data_filename);

    if( (fp = fopen(qsat_data_filename,"r")) == NULL ) {
      ErrPrintf("ERROR: cannot open file %s\n",qsat_data_filename);
      exit(-4);
    }

    //number of points
    if (fscanf(fp,"%d",&nb_rows) != 1){
      ErrPrintf("ERROR: cannot read the number of points in %s\n", qsat_data_filename);
      exit(-4);
    };
    LogPrintf("number of rows: %d\n", nb_rows);

    //read qsat data
    AllocMemory(tau1_values, float, nb_rows);
    AllocMemory(erosion_rates, float, nb_rows);
    AllocMemory(qsat_ratios, float, nb_rows);
    for(i=0; i<nb_rows; i++){
      if (fscanf(fp,"%f %f %f",&tau1_values[i], &erosion_rates[i], &qsat_ratios[i]) != 3){
        ErrPrintf("ERROR: cannot read qsat data in %s (row %d)\n",qsat_data_filename, i);
        exit(-4);
      };
      //LogPrintf("%d\t%f\t%f\t%f\n", i, tau1_values[i], erosion_rates[i], qsat_ratios[i]); //debug
    }

    fclose(fp);
  }

  //computation of the time_scale and tau1
  float ustar; //friction velovity
  float Qsat_ratio; //ratio of sand fluxes
  //float tau1; //threshold of the shear stress
  float qs; //sand flux in the nature
  float erosion_rate; //erosion rate
  float transport_rate; //transport rate
  float deposition_rate; //deposition rate
  float qsat; //sand flux in the model

  if (wind_value >= 0){
    //ustar= wind_value*kappa/log(z/z_0);
    ustar= wind_value*kappa/log(z/z_0);
    Qsat_ratio=1-(ut/ustar)*(ut/ustar);
    for (i=0; (i<nb_rows) && (Qsat_ratio<qsat_ratios[i]); i++);
    if (i<nb_rows){
      //grdvc_min = (tau1_values[i-1]+tau1_values[i])/2; //here we use PDF.data to get tau1 according to the value Qsat_ratio[i]
      //linear interpolation
      float t = (Qsat_ratio - qsat_ratios[i])/(qsat_ratios[i-1] - qsat_ratios[i]);
      grdvc_min = tau1_values[i-1]*t + tau1_values[i]*(1-t); //here we use PDF.data to get tau1 according to the value Qsat_ratio[i]
      erosion_rate = erosion_rates[i];
    }
    else{
      grdvc_min = tau1_values[nb_rows-1];
      erosion_rate = erosion_rates[nb_rows-1];
    }
    if (grdvc_min == 0) grdvc_min = 1.0; //value 0 is not allowed
    grdvc_max = grdvc_min + 1000.0;
    LogPrintf("tau_min = %f\n", grdvc_min);
    //LogPrintf("time_scale = %f\n", time_scale);
  }
  else{
    //if no wind data, then we compute ustar and erosion rate from tau1
    if (grdvc_min == 0) grdvc_min = 1.0; //value 0 is not allowed
    for (i=0; (i<nb_rows) && (grdvc_min>tau1_values[i]); i++);
    Qsat_ratio = qsat_ratios[i];
    erosion_rate = erosion_rates[i];
    if (Qsat_ratio < 1){
      ustar = ut/sqrtf(1 - Qsat_ratio);
    }
    else{
      ustar = 0; //no erosion
    }
  }
  LogPrintf("ustar=%f\n",ustar);
  if (ustar>ut){
    qs = 22.0*(rho_air/rho_sand)*sqrtf(d/g)*(ustar*ustar-ut*ut);  //qsat in the natural
    transport_rate = lambda_T;
    deposition_rate = lambda_C;
    qsat = transport_rate*erosion_rate/deposition_rate; //in model
    time_scale = qsat*(l_0*l_0)/qs; //t_0
  }
  else{
    //no erosion in this case
    //grdvc_min = 1000;
    //grdvc_max = grdvc_min + 1000.0;
    time_scale = 1e6; //not possible to set the time scale accurately
    static char first = 1;
    if (first) {ErrPrintf("Warning: no time scale for wind value %f\n", wind_value); first=0;}
  }
  LogPrintf("time_scale = %g\n", time_scale);
  start = 0;
}
#endif // TIME_SCALE



#ifdef ROTATIONS
void simul_rot()
{
  static int cpt_rot = 0;

  if (flag_save_auto) simul_save_auto();

  //short alea = (drand48()<0.5) ? 1 : 4;
  //rotation(alea*rot_angle, rot_mode);
  if (1) //(rot_angle)
    rotation(rot_angle, rot_mode, 0);
  else
    rotation(360*drand48(), rot_mode, 0);

  cpt_rot++;

  LogPrintf("cpt_rot = %d\n", cpt_rot);

#if defined(GR) && defined(GRJ)
  LogPrintf("grain cells: %d\n", Ncel[GR] + Ncel[GRJ]);
#ifdef AVALANCHES
  if (ava_mode == AVA_SYNC){
    avalanches(GR, ava_h_lim, ava_nb_cel_max, ALTI_MODE_BAS);
  }
#endif
#endif

  flag_csp_altered = 1;

#ifdef LGCA
  flag_lgca_reset = 1;
#endif
}
#endif // ROTATIONS


extern float coef_injection; //regulation de l'injection de grain

void simul_variable_rate()
{
  static char start=1;

#ifdef MODEL_DUN
  extern float lambda_I_per;
  extern float lambda_I_amp;

  if (!lambda_I_amp) return;

  if (start){
    LogPrintf("periodic regulation of the injection rate: period=%f (t0), relative amplitude=%f\n", lambda_I_per, lambda_I_amp);
    start=0;
  }

  // periodic oscillation of the injection rate around the initial value
  coef_injection=1.0+lambda_I_amp*sin(2*PI*csp_time/lambda_I_per);
  Check_min(coef_injection,0);
#endif
}

#ifdef CENTERING_AUTO
void simul_centering()
{
  static char start = 1;
  static int cx0, cz0;

  if (start){
    cx0=roundf(L/2.0 - 0.5);
    cz0=roundf(D/2.0 - 0.5);
    start = 0;
  }

  Vec3 center = compute_mass_center(ALTI);
  int trx = -roundf(center.x - cx0);
  int trz = -roundf(center.z - cz0);
  if (trx || trz){
    translation(trx, trz);
    flag_csp_altered = 1;
#ifdef LGCA
    flag_lgca_reset = 1; /// force flow stabilization
#endif
  }
}
#endif // CENTERING_AUTO


#ifdef SURRECTIONS
void simul_sur()
{
  int h;

  h = compute_alti_max(ALTI);
  LogPrintf("simul_sur: h = %d, csp_time = %f\n", h, csp_time);
  if (h < H-3) surrection();
  else { LogPrintf("surrection bloquee, sommet atteint !\n");  }

  flag_csp_altered = 1;
}
#endif // SURRECTIONS


#ifdef AVALANCHES
/// Perform synchronized avalanches (SYNC mode)
void simul_ava_sync()
{
  avalanches(ALTI, ava_h_lim, ava_nb_cel_max, ALTI_MODE_BAS);

  if (ava_angle){
    avalanches_norm(ALTI, ava_nb_cel_max, ALTI_MODE_BAS);
  }

  compute_normals();

#ifdef DUMP_SIGMA
  dump_sigma_alti();
#endif
#ifdef DUMP_AUTOCORREL
  dump_autocorrel();
#endif

#ifdef LGCA
  lgca_modify_csp();
#endif

  flag_csp_altered = 1;
}


/// Dynamical regulation of the rate and angles of avalanches
/// Typically, we alternate periods without avalanches ("stable" mode) of length ava_delay
/// and periods of avalanches ("unstable" mode) of length ava_duration.
/// More generally, one may define a stable (resp. unstable) rate and a stable (resp. unstable)
/// angle, for black (resp. colored) grains, to simulate complex dynamics based on various angles
/// of repose.
void simul_ava_dynamics_init()
{

#ifdef MODEL_AVA
// valeurs 2D
//  static float ava_delay_short = 500.0;
//  static float ava_delay_long = 4000.0;
// valeurs 3D
//  static float ava_delay_short = 1000.0;
//  static float ava_delay_long = 8000.0;
#else
#ifndef LGCA
// valeurs 2D
//  static float ava_delay_short = 4000.0;
//  static float ava_delay_long = 10000.0;
// valeurs 3D
//  static float ava_delay_short = 1000.0;
//  static float ava_delay_long = 4000.0;
#else
// valeurs 2D
//  static float ava_delay_short = 6000.0;
//  static float ava_delay_long = 12000.0;
// valeurs 3D
//  static float ava_delay_short = 1500.0;
//  static float ava_delay_long = 3000.0;
#endif
#endif

  /*if (ava_mode == AVA_PROPAG){
    time_threshold = 1 + ava_delay*ceil(init_time/ava_delay);
    ava_angle_stable = ava_angle;
    ava_angle_stable_col = ava_angle_col;
    lambda_A_stable = lambda_A;
    //we set the unstable values for the transitions, if not null
    if (ava_angle_unstable) {ava_angle = ava_angle_unstable; mode_stable = 0;}
#ifdef CELL_COLOR
    if (ava_angle_unstable_col) {ava_angle_col = ava_angle_unstable_col; mode_stable = 0;}
#endif
    if (lambda_A_unstable) {lambda_A = lambda_A_unstable; mode_stable = 0;}
    return;
  }*/

  //time_threshold = ava_duration*ceil(csp_time/ava_duration);//*drand48();
  ava_angle_stable = ava_angle;
  lambda_A_stable = lambda_A;
  ava_angle_stable_col = ava_angle_col;
  LogPrintf("simul_ava_dynamics:\n");
  LogPrintf("duration of avalanches = %f\n", ava_duration);
  LogPrintf("delay between avalanches = %f\n", ava_delay);
  LogPrintf("ava_angle = %f (stable)| %f (unstable)\n", ava_angle_stable, ava_angle_unstable);
#ifdef CELL_COLOR
  LogPrintf("ava_angle_col = %f (stable)| %f (unstable)\n", ava_angle_stable_col, ava_angle_unstable_col);
#endif
  //LogPrintf("nb min iterations avant avalanches = %lu\n", inter_iter_ava);
}

double simul_ava_dynamics()
{
  static char mode_stable = 1;
  double next_delay = 0.0;

  if (mode_stable){
    mode_stable = 0;
    ava_angle = ava_angle_unstable;
    lambda_A = lambda_A_unstable;
    next_delay = ava_delay;
  }
  else{
    mode_stable = 1;
    ava_angle = ava_angle_stable;
    lambda_A = lambda_A_stable;
    next_delay = ava_duration;
  }
  //LogPrintf("ava_angle = %f\n",ava_angle);
  //LogPrintf("lambda_A = %f\n",lambda_A);
#ifdef CELL_COLOR
  if (ava_angle_col == ava_angle_stable_col){
    ava_angle_col = ava_angle_unstable_col;
  }
  else{
    ava_angle_col = ava_angle_stable_col;
  }
  //LogPrintf("ava_angle_col = %f\n",ava_angle_col);
#endif //CELL_COLOR

  //LogPrintf("simul_ava_dynamics: next_delay = %f\n", next_delay)
  return next_delay;
}
#endif //AVALANCHES

#ifdef CGV
//extern float lambda_E; //erosion rate
extern float grdvc_max; //seuil max pour le controle de l'erosion en fonction du gradient de vitesse
extern float grdvc_min; //seuil min pour le controle de l'erosion en fonction du gradient de vitesse

/// Random variations of wind duration and strength (erosion threshold)
void simul_variable_wind()
{
  static float delai_vent_max = 200.0; //1000.0; //10.0;
  static float time_threshold = 0.0;
  //static float pente_erosion_grdv = 0.0;
  //static float lambda_E_debut = 0.0;
  static char start = 1;

  if (start){
    time_threshold = delai_vent_max*ceil(csp_time/delai_vent_max); //delai_vent_max*0.5;
    //lambda_E_debut = lambda_E;
    //pente_erosion_grdv = lambda_E/(grdvc_max-grdvc_min);
    //LogPrintf("parametres d'intensite du vent au debut : lambda_E = %f   grdvc_min = %f   grdvc_max = %f\n", lambda_E, grdvc_min, grdvc_max);
    //LogPrintf("pente_erosion_grdv = %f\n", pente_erosion_grdv);
    LogPrintf("parametres d'intensite du vent au debut : tau_min = %f   tau_max = %f\n", grdvc_min, grdvc_max);
    LogPrintf("delai max de variation du vent = %f\n", delai_vent_max);
    start = 0;
  }
  if (csp_time >= time_threshold){
    //lambda_E = 2.0*lambda_E_debut*drand48();
    //lambda_E = lambda_E_debut/(1.1-drand48());
    //grdvc_min = grdvc_max - lambda_E/pente_erosion_grdv;
    grdvc_min = drand48()*500.0;
    grdvc_max = grdvc_min + 1000.0;
    float delai_vent_alea = delai_vent_max*drand48();
    //if (delai_vent_max>=500){ //affichage de l'intensite seulement en frequence basse
      //LogPrintf("variation d'intensite du vent : lambda_E = %f   grdvc_min = %f   csp_time = %f\n", lambda_E, grdvc_min, csp_time);
      LogPrintf("seuils de transport : tau_min = %f   tau_max = %f   csp_time = %f\n", grdvc_min, grdvc_max, csp_time);
      LogPrintf("delai de variation du vent = %f\n", delai_vent_alea);

    //}
    time_threshold += delai_vent_alea;
  }
}
#endif //CGV

#ifdef WIND_DATA
/// Load wind data from an ascii file (periodic sequence of successive winds).
/// Format:
/// nb_wind,unit (on 1st line, where nb_wind is the number of winds in the sequence, and unit is generally "t0")
/// delay angle value (on each subsequent line, with delay in unit, angle in degree, value is the transport threshold Tau_min)
/// ...
/// For real data, set unit to "rt" (standing for real-time).
/// The delay is in seconds, and the value column is the wind speed in m/s (experimental).
double simul_wind_data()
{
  static float *wind_delays = NULL;
  static float *wind_angles = NULL;
  static float *wind_values = NULL;
  static int nb_wind=0;
  static int time_unit=0;
  static int cpt_wind=0;
  static float angle_dir = 0.0;
  double time_threshold = 0.0;
  double delay = 0.0;
  float angle = 0.0;
  //float alpha = 0.0;
  float seq_duration = 0.0;
  char time_unit_str[128];
  static char start = 1;

  if (start){
    LogPrintf("simul_wind_data: read file %s\n", wind_data_filename);
    LogPrintf("rotation mode = %s\n", (rot_mode == ROT_MODE_DISK) ? "rotating table" : (rot_mode == ROT_MODE_FULL) ? "full space" : "overlap");

    //open wind data file
    FILE *fp;
    if( (fp = fopen(wind_data_filename,"r")) == NULL ) {
      ErrPrintf("ERROR: cannot open file %s\n",wind_data_filename);
      exit(-4);
    }

    //number of wind regimes
    if (fscanf(fp,"%d,%s",&nb_wind,time_unit_str) != 2){
      ErrPrintf("ERROR: cannot read the number of wind regimes in %s\n",wind_data_filename);
      exit(-4);
    };
    LogPrintf("number of winds: %d\n",nb_wind);
    //LogPrintf("%s\n",time_unit_str);
    if (!strcmp(time_unit_str,"t0")){
      time_unit = UNIT_T0;
      LogPrintf("time unit is t0\n");
    } else if (!strcmp(time_unit_str,"rt")){
      time_unit = UNIT_RT;
      LogPrintf("time unit is physical\n");
#ifndef TIME_SCALE
      ErrPrintf("ERROR: no time scale, not possible to compute wind data from file %s\n",wind_data_filename);
      exit(-4);
#endif
    } else {
      ErrPrintf("ERROR: cannot read time unit in %s\n",wind_data_filename);
      exit(-4);
    }

    //read wind data
    int i;
    AllocMemory(wind_delays, float, nb_wind);
    AllocMemory(wind_angles, float, nb_wind);
    AllocMemory(wind_values, float, nb_wind);
    for(i=0; i<nb_wind; i++){
      if (fscanf(fp,"%f %f %f",&wind_delays[i], &wind_angles[i], &wind_values[i]) != 3){
        ErrPrintf("ERROR: cannot read wind data in %s\n",wind_data_filename);
        exit(-4);
      };
      //duration of the wind sequence in t0 unit
#ifdef TIME_SCALE
      if (time_unit == UNIT_RT){
        simul_time_scale(wind_values[i]);
        seq_duration += wind_delays[i]/time_scale;
      }
#endif
      if (time_unit == UNIT_T0){
        seq_duration += wind_delays[i];
      }
      if (i<10) LogPrintf("%d\t%f\t%f\t%f\n", i, wind_delays[i], wind_angles[i], wind_values[i]);
    }
    if (i>10) LogPrintf("...\n");
    LogPrintf("duration of the wind sequence = %f t0\n", seq_duration);

    fclose(fp);

    //process wind data
#ifdef TIME_SCALE
    if (time_unit == UNIT_RT){
      //real time
      //initial wind and time_threshold
      time_threshold = seq_duration*floor(csp_time/seq_duration);
      while (time_threshold <= csp_time){
        simul_time_scale(wind_values[cpt_wind]);
        time_threshold += wind_delays[cpt_wind++]/time_scale;
      }
      cpt_wind--;
      LogPrintf("cpt_wind = %d   time_threshold = %f\n", cpt_wind, time_threshold);
      //first wind orientation (north = 0 degree)
      angle = wind_angles[cpt_wind] + 90;
    }
#endif

    if (time_unit == UNIT_T0){
      //virtual time unit t0
      //initial wind and time_threshold
      time_threshold = seq_duration*floor(csp_time/seq_duration);
      while (time_threshold <= csp_time){
        time_threshold += wind_delays[cpt_wind++];
      }
      cpt_wind--;
      LogPrintf("cpt_wind = %d   time_threshold = %f\n", cpt_wind, time_threshold);
      //first wind orientation (west = 0 degree)
      angle = wind_angles[cpt_wind];
#ifdef CGV
      //first value of min shear stress
      if (!param_is_set("Tau_min")) grdvc_min = wind_values[cpt_wind];
#ifdef TIME_SCALE
      //time scale
      simul_time_scale(-1);
#endif
      LogPrintf("tau_min = %f\n", grdvc_min);
#endif
    }

    if (flag_save_auto) simul_save_auto();
    LogPrintf("simul_wind_data: direction = %f, rotation = %f, time = %e\n", wind_angles[cpt_wind], angle, csp_time);
    rotation(-angle, rot_mode, 0);
    angle_dir = wind_angles[cpt_wind]; //first wind direction
    wind_angle_is_set = 1;
    //LogPrintf("simul_wind_data: first direction = %f, time_threshold = %f\n", angle_dir,time_threshold);

#ifdef WIND_ROSE
    simul_wind_rose(nb_wind, wind_angles, wind_delays);
#endif // WIND_ROSE

    delay = time_threshold - csp_time;

    start = 0;
  }
  else{
    //we change the wind counter
    cpt_wind++;
    if (cpt_wind == nb_wind) cpt_wind = 0;
    //wind rotation
    angle = wind_angles[cpt_wind] - angle_dir;
    if (flag_save_auto) simul_save_auto();
    LogPrintf("simul_wind_data: direction = %f, rotation = %f, time = %e\n", wind_angles[cpt_wind], angle, csp_time);
    LogPrintf("cpt_wind = %d\n", cpt_wind);
    //int vol0 = Ncel[GR]+Ncel[GRJ];
    //LogPrintf("volume of sand = %d\n", vol0); //debug
    rotation(-angle, rot_mode, 0);
    //int vol1 = Ncel[GR]+Ncel[GRJ];
    //LogPrintf("volume of sand = %d (%.3f %%)\n", vol1, 100.0*(vol1-vol0)/vol0); //debug
    //new wind direction
    angle_dir = wind_angles[cpt_wind];
#ifdef TIME_SCALE
    if (time_unit == UNIT_RT){
      //time scale
      simul_time_scale(wind_values[cpt_wind]);
      //real time
      delay = wind_delays[cpt_wind]/time_scale;
      LogPrintf("real_time = %e\n", real_time);
    }
#endif

    if (time_unit == UNIT_T0){
      //virtual time unit t0
      delay = wind_delays[cpt_wind];
#ifdef CGV
      //next value of min shear stress
      if (!param_is_set("Tau_min")) grdvc_min = wind_values[cpt_wind];
#ifdef TIME_SCALE
      //time scale
      simul_time_scale(-1);
      LogPrintf("real_time = %e\n", real_time);
#endif
      LogPrintf("tau_min = %f\n", grdvc_min);
#endif
    }
  }

#ifdef LGCA
  if (angle) flag_lgca_reset = 1;
#endif

  if (angle) flag_csp_altered = 1;

  //LogPrintf("simul_wind_data - delay = %f\n", delay);
  return delay;
}
#endif //WIND_DATA

#ifdef WIND_ROSE
/// Compute the distribution of wind angles using wind durations.
/// Wind intensities are not taken into account.
void simul_wind_rose(int nb_winds, float* wind_angles, float* wind_dur)
{
  int step = 10; /// sampling step (in degrees)
  int nb_dir = 360/step; /// max number of wind directions in the rose
  int dir, i;
  float seq_duration = 0;
  float angle;

  LogPrintf("compute wind rose\n");

  /// Initialize wind rose structure
  wind_rose.step = step;
  wind_rose.rstep = step*PI/180.0;
  wind_rose.nb_dir = nb_dir;
  AllocMemory(wind_rose.dist, float, nb_dir);
  ResetMemory(wind_rose.dist, float, nb_dir);

  if (wind_data_filename){
    /// Compute distribution from wind data
    seq_duration = 0;
    for(i=0; i<nb_winds; i++){
      //dir = Round(fmodf(wind_angles[i], 360.0)/step);
      angle = wind_angles[i] + step/2;
      Modulo(angle, 360.0);
      dir = (int)angle/step;
      //LogPrintf("angle=%f, dir=%d,\n",angle,dir);
      Assert((dir>=0)&&(dir<nb_dir));
      wind_rose.dist[dir] += wind_dur[i];
      seq_duration += wind_dur[i];
    }
  }
  else if (rot_delay){
    /// Compute distribution from rot_angle
    seq_duration = 0;
    //nb_winds = (stop_time > csp_time) ? (stop_time-csp_time)/rot_delay : 100;
    nb_winds = 3000;
    angle = 0;
    for(i=0; i<nb_winds; i++, angle+=rot_angle){
      Modulo(angle, 360.0);
      dir = (int)angle/step;
      //LogPrintf("angle=%f, dir=%d,\n",angle,dir);
      Assert((dir>=0)&&(dir<nb_dir));
      wind_rose.dist[dir] += rot_delay;
    }
    seq_duration = nb_winds*rot_delay;
  }
  else return;

  /// Normalize distribution (sum = 1)
  wind_rose.max_val = 0;
  for(dir=0; dir<nb_dir; dir++){
     wind_rose.dist[dir] /= seq_duration;
     wind_rose.max_val = Max(wind_rose.max_val, wind_rose.dist[dir]);
     //LogPrintf("%d-%d: %f\n", dir*step, (dir+1)*step, wind_rose.dist[dir]);
  }
  Assert(wind_rose.max_val > 0);
}
#endif // WIND_ROSE

#ifdef LGCA
extern unsigned char opt_vss;
extern float *grdv;

/// Estimate computing time of lattice gas with respect to stochastic engine
void simul_profile_lgca(int lgca_flag)
{
  static double lgca_elapsed_time = 0.0;
  static double other_elapsed_time = 0.0;
  static double t0 = 0.0;
  static double t1 = 0.0;
  static int cpt=0;

  if (cpt>1000) return;

  if (lgca_flag == 0){
    if (!cpt) LogPrintf("start profiling\n");
    elapsed(&t0);
    if (t1) other_elapsed_time += (t0-t1);
  }
  else if (lgca_flag == 1){
    elapsed(&t1);
    lgca_elapsed_time += (t1-t0);
    if (cpt && (cpt%100==0)){
      LogPrintf("csp_time = %f\n", csp_time);
      if (mode_pat){
        LogPrintf("time spent in lattice gas : %.2f sec (%.1f %%)\n", lgca_elapsed_time, 100*lgca_elapsed_time/(other_elapsed_time + lgca_elapsed_time));
        LogPrintf("time spent waiting for stochastic engine : %.2f sec (%.1f %%)\n", other_elapsed_time, 100*other_elapsed_time/(other_elapsed_time + lgca_elapsed_time));
      }
      else{
        LogPrintf("time spent in stochastic engine : %.2f sec (%.1f %%)\n", other_elapsed_time, 100*other_elapsed_time/(other_elapsed_time + lgca_elapsed_time));
        LogPrintf("time spent in lattice gas : %.2f sec (%.1f %%)\n", lgca_elapsed_time, 100*lgca_elapsed_time/(other_elapsed_time + lgca_elapsed_time));
      }
    }
    cpt++;
    if (cpt>1000) LogPrintf("stop profiling\n");
  }
}

/// Stabilize lattice gas
void flow_stabilization(int nb_cyc)
{
  int i;
  char str_status[100];
  char flag_interp=0;

  LogPrintf("flow stabilization : %d cycles\n", nb_cyc);
  push_status("flow stabilization (0%)",0);
  lgca_ready = 0;

  for (i=0; i<nb_cyc; i++) {
    lgca_collisions();
    lgca_propagations();
    if (lgca_change_visc) lgca_viscosity();
    /// interpolation only during the last cycle (faster)
    flag_interp = (i == nb_cyc-1);
    /*if (i>=nb_cyc-VSTEP_TIME)*/ compute_vel(flag_interp);
    if (!opt_nv){
      sprintf(str_status, "flow stabilization (%d%%)", (100*i)/nb_cyc);
      pop_status(0);
      push_status(str_status,0);
    }
    do_thread_sched();
  }
  flag_lgca_reset = 0;
  lgca_ready = 1;
  LogPrintf("end of flow stabilization : Vsv.meanvel = %f   Vsv.maxvel = %f\n",Vsv.meanvel,Vsv.maxvel);
  pop_status(0);
}

/// Perform one cycle of lattice gas
void simul_lgca()
{
  static int start = 1;

  /// start profiling
  simul_profile_lgca(0);

  /// lgca stabilization
  if (flag_lgca_reset){
    if (start){
      if (!nom_fic_mvt) init_mvt();
      flow_stabilization(init_mvt_nb_cycles);
    }
    else{
      init_mvt();
      flow_stabilization(mvt_nb_cycles);
    }
    start = 0;
  }

  /// lgca cycle
  if (rot_map) out_of_space_mvt(0);
  //LogPrintf("collision ...");
  lgca_collisions();
  //LogPrintf("propagation ...");
  lgca_propagations();
  if (lgca_change_visc) lgca_viscosity();
  //LogPrintf("end of lgca.\n");
  compute_vel(1);

  /// stop profiling
  simul_profile_lgca(1);
}

/// perform full coupling between lattice gas and stochastic engine (DUN model)
void simul_lgca_coupling()
{
#ifdef PARALLEL_AUTOMATA
  if (mode_pat) wait_lgca_thread(); else simul_lgca();
#else
  /// perform one cycle of lattice gas
  simul_lgca();
#endif
#ifdef ALTI
  /// compute normal vectors
  compute_normals();
#endif
#ifdef CGV
  /// compute shear stress
  calcule_grad_vel();
#endif // CGV
}

#endif // LGCA


/// Prepare data and initialize all specific events
/// before the simulation of the current model
void init_simul_models()
{
  LogPrintf("initialization of model specific events and data\n");

#ifdef ROTATIONS
  /// default rotation mode
  if (!param_is_set("Rot_mode")){
    rot_mode = (rot_delay || wind_data_filename)? ROT_MODE_FULL : ROT_MODE_OVERLAP;
  }
#endif

#ifdef WIND_DATA
  if (wind_data_filename){
    schedul_add_non_reg_event(EVT_WIND, csp_time);
  }
#endif

#ifdef TIME_SCALE
  /// time scale computation
  if ((use_lgca) && (!wind_data_filename)) simul_time_scale(-1);
#endif

#ifdef DUMP_SIGMA
  #define SIGMA_DELAY 1.0 //delay in t0 unit
  schedul_add_reg_event(EVT_SIGMA, SIGMA_DELAY, 0);
#endif

#ifdef DUMP_AUTOCORREL
  #define AUTOCOR_DELAY 10.0 //delay in t0 unit
  schedul_add_reg_event(EVT_AUTOCOR, AUTOCOR_DELAY, 0);
#endif

#ifdef CENTERING_AUTO
  //simul_centering();
  if (centering_delay) schedul_add_reg_event(EVT_CENTERING, centering_delay, 0);
#endif


#ifdef ROTATIONS
  if (rot_delay){
    LogPrintf("rotation mode: %s\n", (rot_mode == ROT_MODE_DISK) ? "rotating table" : (rot_mode == ROT_MODE_FULL) ? "full space" : "overlap");
    LogPrintf("rotation delay: %f\n", rot_delay);
#ifdef LGCA
    if (use_lgca){
      LogPrintf("number of lattice-gas stabilization cycles after rotation = %d\n", mvt_nb_cycles);
      if (mvt_nb_cycles < VSTEP_TIME) ErrPrintf("WARNING: stabilization after rotation is too short ( Lgca_nb_cycles = %d < VSTEP_TIME = %d )\n", mvt_nb_cycles, VSTEP_TIME);
    }
#endif
    simul_wind_rose(0, NULL, NULL);
    schedul_add_reg_event(EVT_ROT, rot_delay, 1);
    wind_angle_is_set = 1;
  }
#endif

#ifdef AVALANCHES
  if (ava_mode == AVA_SYNC){
    if (!ava_delay){
      if (lambda_A > 0){
        ava_delay = 1/lambda_A;
      }
      else{
        ErrPrintf("ERROR: bad value (0) for ava_delay\n");
        exit(-1);
      }
    }
    LogPrintf("ava_delay = %f\n", ava_delay);
    LogPrintf("ava_h_lim = %d\n", ava_h_lim);
    LogPrintf("ava_nb_cel_max = %d\n", ava_nb_cel_max);
    schedul_add_reg_event(EVT_AVA_SYNC, ava_delay, 0);
  }

  if ((ava_mode == AVA_TRANS) && ava_delay){
    /// dynamics of avalanches
    simul_ava_dynamics_init();
    schedul_add_non_reg_event(EVT_AVA_DYN, csp_time);
  }
#endif //AVALANCHES

#ifdef LGCA
#endif //LGCA

#ifdef LGCA
  if (use_lgca){
    /// set default number of cycles of stabilization
    if (!param_is_set("Lgca_nb_cycles") && !param_is_set("Init_ncycl")) mvt_nb_cycles = (!lgca_speedup) ? 2*L : 3*VSTEP_TIME;

    /// set initial number of cycles of stabilization
    if (nom_fic_mvt){
      init_mvt_nb_cycles = VSTEP_TIME;
    }
    else{
      init_mvt_nb_cycles = mvt_nb_cycles;
    }

    flag_lgca_reset = 1;

    /// frequency of lattice gas
    if (!lgca_delay){
#ifdef MODEL_DUN
      /// default frequency for DUN model
      lgca_delay = 1.0;
#else
      /// compute frequency from transition rates
      double lambda_max=0;
      for(i=0; i<nb_trans_db; i++){
        if (t_trans[i].intensite > lambda_max) lambda_max = t_trans[i].intensite;
      }
      lgca_delay = 1.0 / lambda_max;
#endif //MODEL_DUN
    }
    LogPrintf("lgca_delay = %f\n", lgca_delay);
    schedul_add_reg_event(EVT_LGCA, lgca_delay, 0);
  }
#endif //LGCA

#ifdef ALTI
  extern unsigned char opt_ls;
  if (!use_lgca && opt_ls){
    /// set regular update of the normals with "ls" option (light source rendering)
    schedul_add_reg_event(EVT_NORMALS, 1.0, 0);
  }
#endif

#ifdef TRACE_FLUX
    extern int use_trace_flux;
    extern float trace_flux_delay;
    if (use_trace_flux && trace_flux_delay){
      schedul_add_reg_event(EVT_SAVE_FLUX, trace_flux_delay, 1);
    }
#endif // TRACE_FLUX
}


void simul_model(int ii)
{
#ifdef CGV
    /// random thresholds of erosion
    //simul_variable_wind();
#endif

#ifdef MODEL_DUN
    /// periodic variability of transition rates (e.g., injection rate)
    simul_variable_rate();
#endif

#if defined(CELL_COLOR) && defined(MODEL_AVA)
   /*if (!(ii & 0x00007fff)){ /// every 65536 transitions
     init_color();
   }*/
#endif

#ifdef TRACE3D_CEL
    //if (!(ii & 0x000001fff)) trace3d_cel();  //toutes les 8192 transitions
    //if (!(ii & 0x00000000)) trace3d_cel();  //toutes les 256 transitions
    trace3d_cel();  /// call function at every step
#endif
}


double simul_model_event(int type)
{
  double delay = 0.0;

  switch (type){
#ifdef LGCA
    case EVT_LGCA:
      simul_lgca_coupling();
      break;
#endif // LGCA

#ifdef ALTI
    case EVT_NORMALS:
      compute_normals();
      break;
#endif // ALTI

#ifdef CENTERING_AUTO
    case EVT_CENTERING:
      simul_centering();
      break;
#endif // CENTERING_AUTO

#ifdef SURRECTIONS
    case EVT_SUR:
      simul_sur();
      break;
#endif // SURRECTIONS

#ifdef ROTATIONS
    case EVT_ROT:
      simul_rot();
      break;
#endif // ROTATIONS

#ifdef AVALANCHES
    case EVT_AVA_SYNC:
      simul_ava_sync();
      break;

    case EVT_AVA_DYN:
      delay = simul_ava_dynamics();
      break;
#endif // AVALANCHES

#ifdef WIND_DATA
    case EVT_WIND:
      delay = simul_wind_data();
      break;
#endif // WIND_DATA

#ifdef DUMP_SIGMA
    case EVT_SIGMA:
      dump_sigma_alti();
      break;
#endif // DUMP_SIGMA

#ifdef DUMP_AUTOCORREL
    case EVT_AUTOCOR:
      dump_autocorrel();
      break;
#endif // DUMP_AUTOCORREL

#ifdef TRACE_FLUX
    case EVT_SAVE_FLUX:
      trace_flux_dump();
      break;
#endif // TRACE_FLUX

    default:
      ErrPrintf("ERROR: event %d not supported in family SIMUL\n", type);
      exit(-1);
  }

  return delay;
}
